prefix={{ prefix }}
exec_prefix=${prefix}
libdir=${prefix}/{{ settings.libraries_prefix }}
includedir=${prefix}/{{ settings.includes_prefix }}


Name: {{ lib.name }}
Description: {{ lib.description }}
Version: {{ lib.version }}
Libs:{% for opt in libs %} {{ opt }}{% endfor %}
Cflags:{% for opt in cflags %} {{ opt }}{% endfor %}
Requires:{% for req in requires %} {{ req }}{% endfor %}
